//------------------------------------------------------------------------
//
// OpenDCC - S88-N - IR Reflex Feedback Module
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      ir_sensor.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-02-18 V0.1 start
//            2008-03-31 V0.2 AREF changed to AVCC
//
//------------------------------------------------------------------------
//
// purpose:   IR feedback system
//
// content:   Evaluation of IR sensor with A/D converter
//
// interrupts:  USI_OVERFLOW (for s88 handling)
//              TIMER_COMPA (for timetick)
//
// timing:    Conversion clk: 1MHz (reduced resolution)
//            each conversion: 13us
//            CNY does a real slow settling - see ..._DELAY
//
// hardware:  this measures the output voltage of the IR transistor;
//            use a pull up to 2,5V to have full adc-range available.
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "main.h"


#define SIMULATION  0            // 0: real application
                                 // 1: ???

#define CNY70_ON_DELAY      500  // time in s, until a measurement could be performed.
#define CNY70_OFF_DELAY     500  // time in s, until a idle state is reached on CNY70
#define ADC_MUX_DELAY         5  // time in s to allow for settling of the adc


//---------------------------------------------------------------------------------------
// ADC lib
//

#define ADC_START  (ADCSR |= (1 << ADSC))
#define ADC_DONE (!(ADCSR & (1 << ADSC)))    // done, if start conversions goes 0

void init_adc(void)
  {
    // Setup Clock
    // AVR runs at 8 MHz, we want 1MHz (fast conversions)
    //
    ADCSR  = (1 << ADEN)    //  (ADC Enable)
           | (0 << ADSC)    //  (ADC Start Conversion)
           | (0 << ADFR)    //  (ADC Free run)
           | (0 << ADIF)    //  (ADC Interrupt Flag)
           | (0 << ADIE)    //  (ADC Interrupt Enable)
           | (0 << ADPS2)   //  0,1,1 = Divider 8
           | (1 << ADPS1)
           | (1 << ADPS0);

    ADMUX = (0 << REFS1)    // 0,0 = AVCC
          | (0 << REFS0)    // 0,1 = AREF external
                            // 1,0 = internal 2.56V no C
                            // 1,1 = internal 2.56V + ext. C
          | (1 << ADLAR)    // 1 = left adjusted
          | (0 << MUX4)     // = Channel
          | (0 << MUX3)
          | (0 << MUX2)
          | (0 << MUX1)
          | (0 << MUX0);

    // start a dummy conversion to initialize ADC

    ADC_START;
    while(!ADC_DONE);
  }
  
unsigned char do_fast_adc(unsigned char channel)
  {
    ADMUX = (0 << REFS1)    // 0,0 = AVCC
          | (0 << REFS0)    // 0,1 = AREF external
                            // 1,0 = internal 2.56V no C
                            // 1,1 = internal 2.56V + ext. C
          | (1 << ADLAR)    // 1 = left adjusted
          | channel;

    ADC_START;

    while(!ADC_DONE);

    return(ADCH);           // only 8 bits - return(ADCW);
    
  }

//-------------------------------------------------------------------------
// we do: 5 measurements
//   LED_OFF:  value A(1)
//   LED_ON:   value B(1)
//   LED_OFF:  value A(2)
//   LED_ON:   value B(2)
//   LED_OFF:  value A(3)
//
// if an object is present: the values for B will be smaller
//
// now calculate: mean[A(i)] >  mean[B(i)] * ratio  
// If this is true, we have an object in front of the sensor
//
// Formula: ratio = 2/3*GAIN_B/GAIN_A
// rearrange this formula, for different ratios we get:
// ratio = 1.0   ->   2 * sum(A(i)) >  3 * sum(B(i))
// ratio = 1.042 ->  16 * sum(A(i)) > 25 * sum(B(i))
// ratio = 1.05  ->  40 * sum(A(i)) > 63 * sum(B(i))
// ratio = 1.066 ->  15 * sum(A(i)) > 24 * sum(B(i)) 
// ratio = 1.083 ->   8 * sum(A(i)) > 13 * sum(B(i)) <<<<
// ratio = 1.1   ->  20 * sum(A(i)) > 33 * sum(B(i))
// ratio = 1.2   ->  10 * sum(A(i)) > 18 * sum(B(i))
// ratio = 1.25  ->   8 * sum(A(i)) > 15 * sum(B(i))
// ratio = 1.33  ->   1 * sum(A(i)) >  2 * sum(B(i))
// ratio = 1.5   ->   4 * sum(A(i)) >  9 * sum(B(i))
#define GAIN_A  8
#define GAIN_B  13

unsigned char read_sensor(unsigned char channel)
  {
    unsigned int A, B;
    IR_LED_OFF;
    _mydelay_us(CNY70_OFF_DELAY);
    A = do_fast_adc(channel);

    IR_LED_ON;
    _mydelay_us(CNY70_ON_DELAY);
    B = do_fast_adc(channel);

    IR_LED_OFF;
    _mydelay_us(CNY70_OFF_DELAY);
    A += do_fast_adc(channel);
    
    IR_LED_ON;
    _mydelay_us(CNY70_ON_DELAY);
    B += do_fast_adc(channel);
    
    IR_LED_OFF;
    _mydelay_us(CNY70_OFF_DELAY);
    A += do_fast_adc(channel);

    A = A * GAIN_A;
    B = B * GAIN_B;

    if (A>B) return(TRUE);
    else return(FALSE);

  }


unsigned char read_all_sensors(void)
  {
    unsigned char channel;
    unsigned char result = 0;
    unsigned int A[8], B[8];
    
    IR_LED_OFF;
    _mydelay_us(CNY70_OFF_DELAY);                               // wait some time to settle CNY70
    for (channel = 0; channel < 8; channel++)
      {
        A[channel] = do_fast_adc(channel);
        _mydelay_us(ADC_MUX_DELAY); 
      }
    IR_LED_ON;
    _mydelay_us(CNY70_ON_DELAY);                               // wait some time to settle CNY70
    for (channel = 0; channel < 8; channel++)
      {
        B[channel] = do_fast_adc(channel);
        _mydelay_us(ADC_MUX_DELAY); 
      }
    IR_LED_OFF;
    _mydelay_us(CNY70_OFF_DELAY);                               // wait some time to settle CNY70
    for (channel = 0; channel < 8; channel++)
      {
        A[channel] += do_fast_adc(channel);
        _mydelay_us(ADC_MUX_DELAY); 
      }
    IR_LED_ON;
    _mydelay_us(CNY70_ON_DELAY);                               // wait some time to settle CNY70
    for (channel = 0; channel < 8; channel++)
      {
        B[channel] += do_fast_adc(channel);
        _mydelay_us(ADC_MUX_DELAY); 
      }
    IR_LED_OFF;
    _mydelay_us(CNY70_OFF_DELAY);                               // wait some time to settle CNY70
    for (channel = 0; channel < 8; channel++)
      {
        A[channel] += do_fast_adc(channel);
        _mydelay_us(ADC_MUX_DELAY); 
      }

    unsigned char mask = 0x80;
    for (channel = 0; channel < 8; channel++)
      {
        A[channel] = A[channel] * GAIN_A;
        B[channel] = B[channel] * GAIN_B;
        
        if (A[channel] > B[channel]) result |= mask;            // s88 has invers order - we do the
                                                                // inversion here
        mask >>= 1;
      }

    return(result);
  }


//=========================================================================





//=========================================================================
//
// Simulationen:
//
// 0: keine Simulation
//
#if (SIMULATION == 1)

#endif // (SIMULATION == 1)
