//------------------------------------------------------------------------
//
// OpenDCC - S88-N - Keyboard and RC5 Receiver
//
// Copyright (c) 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      s88_key.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-11-27 V0.1 start
//            2008-01-10 V0.2 keyboard tested and okay (both modes)
//            2008-03-18 V0.3 copied from s88 keyboard, here only debounce used
//
//------------------------------------------------------------------------
//
// purpose:   keyboard routines for directly connected keys
//
// content:   Here we scan the inputs and do the debounce.
//
//            Every Input has a holdtime - this will prolong
//            each stroke to a given time.
//
//            Currently we support two types of keyboards
//              1: standard direct keys to ground.
 //             2: Pollin Dual Scan (Samwell with 12 Keys) 
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "main.h"                // timerval, current_message
//------------------------------------------------------------------------------

#define KEY_SCAN_TICK    8000L    // scan keyboard every 10ms
#define KEY_DEBOUNCE     (20000L / KEY_SCAN_TICK)        // 20ms Entprellen
#define HOLDTIME         (1000000L / KEY_SCAN_TICK)      // Hold 1s

#if (HOLDTIME > 255)
#warning HOLDTIME Overflow, please use a small value or larger SCAN_TICK!
#endif

#define nop() __asm volatile ("nop")

//==============================================================================
//
#define NUM_OF_KEYS      8    // 12

struct
  {
    enum                  
      {                           
        S_UNPRESSED,
        S_PRESS_DEBOUNCE,
        S_PRESSED,
        S_RELEASE_DEBOUNCE
      } state; 
    unsigned char holdtime;   // time to keep the pressed state (given in ticks (10ms))
    unsigned char debounce;   // debounce in both directions
  } key[NUM_OF_KEYS];


//========================================================================================
//
//  Keyboard - DeBounce - Handler
//
//========================================================================================

void init_debounce(void)
  {
    unsigned char i;

    for (i=0; i<NUM_OF_KEYS; i++)
      {
        key[i].state = S_UNPRESSED;
        key[i].holdtime = 0;
      }
  }


unsigned char  debounce(unsigned char pattern)
  {
    unsigned char mask;
    unsigned char debounced = 0;
    
    unsigned char i;

    // pattern = read_keys();             // fetch actual state
    

    current_message = 0;                // auto reset
    mask = 1;
    for (i=0; i<NUM_OF_KEYS; i++)
      {
        switch(key[i].state)
          {
            case S_UNPRESSED:
                if ((pattern & mask))
                  {
                    // got pressed
                    key[i].debounce = KEY_DEBOUNCE;
                    key[i].state = S_PRESS_DEBOUNCE;
                  }
                break;
            case S_PRESS_DEBOUNCE:
                if ((pattern & mask))
                  {
                    // still pressed
                    key[i].debounce--;
                    if (key[i].debounce == 0) key[i].state = S_PRESSED;
                  }
                else
                  {
                    key[i].state = S_UNPRESSED;
                  }
                break;
            case S_PRESSED:
                if ((pattern & mask))
                  {
                    // still pressed
                    key[i].holdtime = HOLDTIME;
                  }
                else
                  {
                    // got released
                    key[i].debounce = KEY_DEBOUNCE;
                    key[i].state = S_RELEASE_DEBOUNCE;
                  }
                break;
            case S_RELEASE_DEBOUNCE:
                if ((pattern & mask))
                  {
                    // still pressed - go back 
                    key[i].state = S_PRESSED;
                  }
                else
                  {
                    
                    key[i].debounce--;
                    if (key[i].debounce == 0) key[i].state = S_UNPRESSED;
                  }
                break;
          }

        if (key[i].holdtime > 0)                // whoever has a holdtime > 0: is active
          {
            key[i].holdtime--;
            debounced |= mask;
          }
        mask = mask << 1;
      }
    return(debounced);
  }


