//------------------------------------------------------------------------
//
// OpenDCC - S88-N - IR Reflex Feedback Module
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer (main)
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-02-16 V0.1 start
//            2008-05-19 V0.2 Timer1 init 
//
//------------------------------------------------------------------------
//
// purpose:   IR feedback system
//
// content:     1. Defines and variable definitions
//              2. Hardware init
//              3. s88 handle
//              4. main
//
// interrupts:  USI_OVERFLOW (for s88 handling)
//              TIMER1_OVF1 (for timetick)
//
// timing:    on s88:
//            from falling clk edge until update of port:
//               t-min: 10us;
//               t-max: 15us (with extra sei() in RC5 interrupt routine)
//               t-max: 25us (normal interrupt routine)
//               these values are not visible, since we shift out data
//               by hardware.
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"             // system configuration
#include "ir_sensor.h"          // read in cny70
#include "key_debounce.h"       // generic debounce algorithm
#include "main.h"


#define SIMULATION  0            // 0: real application
                                 // 1: ???

#define SHOW_S88_ON_LED        TRUE    // TRUE: turn on LED each s88 clk


#define DIRECT_READ  0            // 0: real application - detectors
                                  // 1: read channel 0 and send to s88 as is
                                  // 2: fixed to 0x7A - testing s88 only

//
//-------------------------------------------------------------------------
//
// Connect the Attiny26 as follows:
//
//                       --------U--------
//               S88_DI  | 1 PB0   PA0 20|- ADC0 - IR
//               S88_DO  | 2 PB1   PA1 19|- ADC1 - IR
//            S88_CLOCK  | 3 PB2   PA2 18|- ADC2 - IR
//         MOSFET IR_ON  | 4 PB3   PA3 17|- AREF ----||--- GND
//                  +5V  | 5 VCC   GND 16|- GND
//                  GND  | 6 GND   VCC 15|- VCC
//    (xtal1) ADC7 - IR  | 7 PB4   PA4 14|- ADC3 - IR
//   +--|>|-- 2k -- LED  | 8 PB5   PA5 13|- ADC4 - IR
//             S88_LOAD  | 9 PB6   PA6 12|- ADC5 - IR
//                RESET  |10 RST   PA7 11|- ADC6 - IR
//                       -----------------
//
// Fuses:
//             7 6 5 4 3 2 1 0
//  high byte: 1 1 1 1 0 1 0 0
//             | | | | | | | |- BODEN (0=Enable BOD)
//             | | | | | | |--- BODLEVEL (0=BOD at 4.0)
//             | | | | | |----- EESAVE (1=do not keep eeprom)
//             | | | | |------- SPIEN (0=spi enabled)
//             | | | |--------- RSTDISBL (1=Reset as is)
//             | | |----------- reserved
//             | |------------- reserved
//             |--------------- reserved
//
//   low byte: 1 1 0 0 0 1 0 0
//             | | | | | | | |- CKSEL0: (1=standard boot)
//             | | | | | | |--- CKSEL1
//             | | | | | |----- CKSEL2
//             | | | | |------- CKSEL3 (0100 = RC, 8 MHz)
//             | | | |--------- SUT0 
//             | | |----------- SUT1 (00=bod)
//             | |------------- CKOPT (1=no 36pF)
//             |--------------- PLL (1=do not divide)
//
//----------------------------------------------------------------------------


volatile signed char timerval;          // is incremented by ISR

volatile unsigned char Communicate;     // bitfield for interprocess communication

volatile unsigned int current_message;  // this is the s88-message


void init_main(void)
  {
                                 // PortA: unchanged
    DDRB  =  0                   // PortB: 1=out, 0=in
          | (1 << IR_LED)
          | (1 << LED)          
          | (0 << PS88_CLK)  
          | (1 << PS88_DO)
          | (0 << PS88_DI)
          | (0 << PS88_LOAD);
    PORTB  = 0                   // PortB: 1=pull up
          | (1 << IR_LED)
          | (1 << LED)          // turn off LED          
          | (1 << PS88_CLK)  
          | (0 << PS88_DO)
          | (1 << PS88_DI)
          | (1 << PS88_LOAD);

    
    // Init Timer1 as CTC, first do a check on TICK_PERIOD and F_CPU
    #define TIMER1_CLKDIV 1024              // possible values: 1 ... 16384 (2^^n)

    #if (F_CPU / 1000000L * TICK_PERIOD / TIMER1_CLKDIV) > 255L
      #warning: overflow in OCR1C - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a larger clkdiv
    #endif    
    #if (F_CPU / 1000000L * TICK_PERIOD / TIMER1_CLKDIV) < 60L
      #warning: resolution accuracy in OCR1C too low - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a smaller clkdiv
    #endif    


    // OCR1A = F_CPU / 1000000L * TICK_PERIOD / TIMER1_CLKDIV ;  
    OCR1C = F_CPU / 1000000L * TICK_PERIOD / TIMER1_CLKDIV ;   // clear at compare to c 
    TCCR1A = (0 << COM1A1)      // compare match A
           | (0 << COM1A0) 
           | (0 << COM1B1)      // compare match B
           | (0 << COM1B0) 
           | (0 << FOC1A) 
           | (0 << FOC1B) 
           | (0 << PWM1A)
           | (0 << PWM1B);      // Timer1 
    TCCR1B = 0
           | (1 << CTC1)
           | (0 << PSR1)
           | 0                  // reserved
    #if   TIMER1_CLKDIV == 1
           | (0 << CS13) | (0 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[3:0]=0001 clkdiv 1
    #elif TIMER1_CLKDIV == 2
           | (0 << CS13) | (0 << CS12) | (1 << CS11)  | (0 << CS10);       // CS[3:0]=0010 clkdiv 2
    #elif TIMER1_CLKDIV == 4
           | (0 << CS13) | (0 << CS12) | (1 << CS11)  | (1 << CS10);       // CS[3:0]=0011 clkdiv 4
    #elif TIMER1_CLKDIV == 8
           | (0 << CS13) | (1 << CS12) | (0 << CS11)  | (0 << CS10);       // CS[3:0]=0100 clkdiv 8
    #elif TIMER1_CLKDIV == 16
           | (0 << CS13) | (1 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[3:0]=0101 clkdiv 16
    #elif TIMER1_CLKDIV == 32
           | (0 << CS13) | (1 << CS12) | (1 << CS11)  | (0 << CS10);       // CS[3:0]=0110 clkdiv 32
    #elif TIMER1_CLKDIV == 64
           | (0 << CS13) | (1 << CS12) | (1 << CS11)  | (1 << CS10);       // CS[3:0]=0111 clkdiv 64
    #elif TIMER1_CLKDIV == 128
           | (1 << CS13) | (0 << CS12) | (0 << CS11)  | (0 << CS10);       // CS[3:0]=1000 clkdiv 128
    #elif TIMER1_CLKDIV == 256
           | (1 << CS13) | (0 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[3:0]=1001 clkdiv 256
    #elif TIMER1_CLKDIV == 512
           | (1 << CS13) | (0 << CS12) | (1 << CS11)  | (0 << CS10);       // CS[3:0]=1010 clkdiv 512
    #elif TIMER1_CLKDIV == 1024
           | (1 << CS13) | (0 << CS12) | (1 << CS11)  | (1 << CS10);       // CS[3:0]=1011 clkdiv 1024
    #elif TIMER1_CLKDIV == 2048
           | (1 << CS13) | (1 << CS12) | (0 << CS11)  | (0 << CS10);       // CS[3:0]=1100 clkdiv 2048
    #elif TIMER1_CLKDIV == 4096
           | (1 << CS13) | (1 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[3:0]=1101 clkdiv 4096
    #elif TIMER1_CLKDIV == 8192
           | (1 << CS13) | (1 << CS12) | (1 << CS11)  | (0 << CS10);       // CS[3:0]=1110 clkdiv 8192
    #elif TIMER1_CLKDIV == 16384
           | (1 << CS13) | (1 << CS12) | (1 << CS11)  | (1 << CS10);       // CS[3:0]=1111 clkdiv 16384
    #else
     #warning: TIMER1_CLKDIV is void
    #endif

    TIMSK = (1<<TOIE1)       // Timer1 Overflow (used here)
          | (0<<OCIE1A)      // Timer1 Compare A
          | (0<<OCIE1B)      // Timer1 Compare B
          | 0                // reserved
          | 0                // reserved
          | (0<<TOIE0)       // Timer0 Overflow
          | 0;                // reserved


    // Init Interrupt 0

    GIMSK = (0<<INT0)        
          | (0<<PCIE0)        
          | (0<<PCIE1); 

    MCUCR = 0x00;  //           // no interrupts ;The rising edge of INT0 generates an interrupt request.


    // USI

    USICR = (0 << USISIE)    // Start Condition Interrupt Enable   
          | (1 << USIOIE)    // Counter Overflow Interrupt Enable       
          | (0 << USIWM1)    // USIWM1..0: Wire Mode     
          | (1 << USIWM0)    // 00: normal Mode
                             // 01: three wire mode (used here)
                             // 10: two wire mode
          | (1 << USICS1)    // Clock Source:    
          | (0 << USICS0)    // 10: external, pos. edge    
          | (0 << USICLK)    // Clkout (not used here)
          | (0 << USITC);    // Toggle     
          
         
    USISR = 0b00001111;      // set Counter to 15, to get an interrupt on next clk change   
  }


unsigned char copy[] PROGMEM = {".... S88_IR_Reflex V0.1 (c) Kufer 2008 ..."};


//---------------------------------------------------------------------------------------------

ISR(TIMER1_OVF1_vect)
  {
    timerval++;
  }


//===============================================================================

// Long Bit fifo
//
// Put new bit to LSB of buffer[0], then shift complete buffer up one bit.
// Return top byte of Buffer.
// Note: declare buffer volatile, not to disturb the compiler optimization
//       The pointer to the buffer will be destroyed (I don't know how to tell it the
//       compiler), so multiple call with const pointers will dramatically fail :-(
//       (No problem, if used inside a ISR)
//
unsigned char bit_fifo(volatile uint8_t *buffer, uint8_t length, uint8_t newbit)
     __attribute__((always_inline));
unsigned char bit_fifo(volatile uint8_t *buffer, uint8_t length, uint8_t newbit)
  {
      uint8_t shifted;
	
      asm volatile (
         "ror %3"                          "\n"
         "BIT_FF_LOOP%=:"                  "\n"
         "ld %0,  %a1"                     "\n"
         "adc %0, %0"                      "\n"
         "st %a1+, %0"                     "\n"
         "dec %2"                          "\n"
         "brne BIT_FF_LOOP%="              "\n"
            : "=&r" (shifted)                               // outputs
            : "e"  (buffer), "r" (length), "r" (newbit)     // inputs
            : "memory"                                      // clobbered
      );
    return(shifted);
   }




//-----------------------------------------------------------------------------
// S88-Handler
//
// we use USI
// 
// a) 8-Bit mode (shift_size = 8)
//    USI could be used directly
//    states: 1. Shifting (no interaction)
//            2. Loading -> enable int on rising edge
//                          ISR will copy current_message to shifter
//
// b) 16-Bit mode (shift_size = 8..16)
//    USI is too short, needs a software extension
//    In this mode, we use USI only for two msb bit.
//    Interrupt on every clock: rising: 1. Shifting (sample data in)
//                                      2. Loading (copy current_message)
//                              falling: put out next Data
//                                       + prepare next bit (so copy 2 MSBs)
//                                       if there is only the MSB, Data Out
//                                       will have low immediately after clk low
//                                       (you can see timing then)
// S88 Reset is ignored
//-----------------------------------------------------------------------------


unsigned int shift_reg;
unsigned char shift_size = 8;   // may be 8 ... 16
// #define shift_size 16

// case b) Extension of shiftregister in SW
ISR(USI_OVF_vect)          // USI Overflow
  {
    unsigned char my_usi, temp;
    
    #if (SHOW_S88_ON_LED == TRUE)
      LED_ON;
    #endif

    USISR |= 0b00001111;      // set Counter to 15, to get an new interrupt

    #if (SHOW_S88_ON_LED == TRUE)
          if (!semaphor_query(C_LED)) LED_OFF;
    #endif


    if (S88_CLK)
      {
        // we got a rising edge:
        if (S88_LOAD)
          {
            shift_reg = current_message;
            semaphor_set(C_S88_Sent);           // tell main that data is sent

            my_usi = shift_reg >> (shift_size-8);
            // my_usi &= 0x80;  // if only MSB is written, you can observe timing on a scope
            USIDR = my_usi;
          }
        else
          { 
            temp = USIDR & 0x01;                // read input
            shift_reg <<= 1;
            shift_reg += temp;
          }
      }
    else
      {
        // we got a falling edge, write next bit

         my_usi = shift_reg >> (shift_size-8);     // note: shift_size must be >= 8

         //x my_usi = shift_reg >> (shift_size-2);  //x: then it could be used for any size from 2..16
         //x my_usi = my_usi << 6;
         // my_usi &= 0x80;  // if only MSB is written, you can observe timing on a scope
         USIDR = my_usi;
      }
  }




// volatile unsigned char aaa[]={0x01, 0xFF, 0x03, 0xAA, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
// volatile unsigned char *aptr;

int main(void)
  {
    signed char key_timer;
    unsigned char pattern = 0;
    
    init_main();

    init_adc();

    init_debounce();

    shift_size = 8;                            // std keyboard

    sei();                                     // Global enable interrupts

    key_timer = timerval;

    while(1)
      {
        // signed char temp;                   // trick down gcc: this forces char and smaller code
        // temp = timerval - key_timer;
        if (timerval != key_timer)
          {
            key_timer = timerval;
            
            #if (DIRECT_READ == 2)
                #warning DIRECT_READ is set to 2, we send back 0x7A 
                current_message = 0x7a;

            #elif (DIRECT_READ == 1)
                #warning DIRECT_READ is set to 1, we send back the measurement value of channel 0 as is.
                IR_LED_ON;

                _mydelay_us(500);
                current_message = do_fast_adc(0);
                IR_LED_OFF;
            
                if (current_message < 100)
                  {
                    semaphor_set(C_LED);       // set flag for LED active
                    LED_ON;
                  }
                else
                  {
                    semaphor_get(C_LED);       // reset flag for LED active
                    LED_OFF;
                  }

            #else //  (DIRECT_READ == 1)
                                                // normal operation
                pattern = read_all_sensors();
                current_message = debounce(pattern);
                
                if (current_message)
                  {
                    semaphor_set(C_LED);       // set flag for LED active
                    LED_ON;
                  }
                else
                  {
                    semaphor_get(C_LED);       // reset flag for LED active
                    LED_OFF;
                  }
            #endif // (DIRECT_READ == 1)
              
          }               
      }
  }


//=========================================================================





//=========================================================================
//
// Simulationen:
//
// 0: keine Simulation
//
#if (SIMULATION == 1)
#warning SIMULATION on, do not use on real hardware
#endif // (SIMULATION == 1)
